@@ -25,6 +25,31 @@
 /**    OR PERFORMANCE OF THIS SOFTWARE.                                     **/
 /*****************************************************************************/
 
+/* 
+ *  [ ctwm ]
+ *
+ *  Copyright 1992 Claude Lecommandeur.
+ *            
+ * Permission to use, copy, modify  and distribute this software  [ctwm] and
+ * its documentation for any purpose is hereby granted without fee, provided
+ * that the above  copyright notice appear  in all copies and that both that
+ * copyright notice and this permission notice appear in supporting documen-
+ * tation, and that the name of  Claude Lecommandeur not be used in adverti-
+ * sing or  publicity  pertaining to  distribution of  the software  without
+ * specific, written prior permission. Claude Lecommandeur make no represen-
+ * tations  about the suitability  of this software  for any purpose.  It is
+ * provided "as is" without express or implied warranty.
+ *
+ * Claude Lecommandeur DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * INCLUDING ALL  IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS.  IN NO
+ * EVENT SHALL  Claude Lecommandeur  BE LIABLE FOR ANY SPECIAL,  INDIRECT OR
+ * CONSEQUENTIAL  DAMAGES OR ANY  DAMAGES WHATSOEVER  RESULTING FROM LOSS OF
+ * USE, DATA  OR PROFITS,  WHETHER IN AN ACTION  OF CONTRACT,  NEGLIGENCE OR
+ * OTHER  TORTIOUS ACTION,  ARISING OUT OF OR IN  CONNECTION WITH THE USE OR
+ * PERFORMANCE OF THIS SOFTWARE.
+ *
+ * Author:  Claude Lecommandeur [ lecom@sic.epfl.ch ][ April 1992 ]
+ */
 
 /**********************************************************************
  *
@@ -43,8 +68,13 @@
  **********************************************************************/
 
 #include <stdio.h>
+#include <string.h>
 #include "twm.h"
+#ifdef VMS
+#include <decw$include/Xatom.h>
+#else
 #include <X11/Xatom.h>
+#endif
 #include "add_window.h"
 #include "util.h"
 #include "resize.h"
@@ -58,9 +88,9 @@
 
 #define gray_width 2
 #define gray_height 2
-static char gray_bits[] = {
+static unsigned char gray_bits[] = {
    0x02, 0x01};
-static char black_bits[] = {
+static unsigned char black_bits[] = {
    0xFF, 0xFF};
 
 int AddingX;
@@ -71,10 +101,13 @@ int AddingH;
 static int PlaceX = 50;
 static int PlaceY = 50;
 static void CreateWindowTitlebarButtons();
+void ClickToFocusGrab   ();
+void ClickToFocusUngrab ();
 
 char NoName[] = "Untitled"; /* name if no name is specified */
 int  resizeWhenAdd;
 
+extern Atom _OL_WIN_ATTR;
 
 /************************************************************************
  *
@@ -157,7 +190,6 @@ IconMgr *iconp;
 #ifdef DEBUG
     fprintf(stderr, "AddWindow: w = 0x%x\n", w);
 #endif
-
     /* allocate space for the twm window */
     tmp_win = (TwmWindow *)calloc(1, sizeof(TwmWindow));
     if (tmp_win == 0)
@@ -188,6 +220,8 @@ IconMgr *iconp;
       tmp_win->attr.height = Scr->MaxWindowHeight;
 
     tmp_win->wmhints = XGetWMHints(dpy, tmp_win->w);
+    if (tmp_win->wmhints && !(tmp_win->wmhints->flags & InputHint))
+	tmp_win->wmhints->input = True;
     if (tmp_win->wmhints && (tmp_win->wmhints->flags & WindowGroupHint)) 
       tmp_win->group = tmp_win->wmhints->window_group;
     else
@@ -211,19 +245,22 @@ IconMgr *iconp;
     namelen = strlen (tmp_win->name);
 
     tmp_win->highlight = Scr->Highlight && 
-	(!(short)(int) LookInList(Scr->NoHighlight, tmp_win->full_name, 
+	(!LookInList(Scr->NoHighlight, tmp_win->full_name, 
 	    &tmp_win->class));
 
     tmp_win->stackmode = Scr->StackMode &&
-	(!(short)(int) LookInList(Scr->NoStackModeL, tmp_win->full_name, 
+	(!LookInList(Scr->NoStackModeL, tmp_win->full_name, 
 	    &tmp_win->class));
 
+    tmp_win->ontoppriority = (LookInList(Scr->AlwaysOnTopL,
+	tmp_win->full_name, &tmp_win->class)) ? ONTOP_MAX : ONTOP_DEFAULT;
+
     tmp_win->titlehighlight = Scr->TitleHighlight && 
 	(!(short)(int) LookInList(Scr->NoTitleHighlight, tmp_win->full_name, 
 	    &tmp_win->class));
 
     tmp_win->auto_raise = Scr->AutoRaiseDefault ||
-      (short)(int) LookInList(Scr->AutoRaise, tmp_win->full_name,
+	LookInList(Scr->AutoRaise, tmp_win->full_name,
 			      &tmp_win->class);
     if (tmp_win->auto_raise) Scr->NumAutoRaises++;
     tmp_win->iconify_by_unmapping = Scr->IconifyByUnmapping;
@@ -237,7 +274,8 @@ IconMgr *iconp;
 	(short)(int) LookInList(Scr->IconifyByUn, tmp_win->full_name,
 	    &tmp_win->class);
 
-    if (LookInList(Scr->WindowRingL, tmp_win->full_name, &tmp_win->class)) {
+    if (Scr->WindowRingAll ||
+	LookInList(Scr->WindowRingL, tmp_win->full_name, &tmp_win->class)) {
 	if (Scr->Ring) {
 	    tmp_win->ring.next = Scr->Ring->ring.next;
 	    if (Scr->Ring->ring.next->ring.prev)
@@ -272,6 +310,17 @@ IconMgr *iconp;
 
     tmp_win->old_bw = tmp_win->attr.border_width;
 
+    tmp_win->frame_bw3D = Scr->ThreeDBorderWidth;
+    if (LookInList(Scr->NoBorder, tmp_win->full_name, &tmp_win->class)) {
+	tmp_win->frame_bw = 0;
+	tmp_win->frame_bw3D = 0;
+    }
+    else
+    if (tmp_win->frame_bw3D != 0) {
+	tmp_win->frame_bw = 0;
+	Scr->ClientBorderWidth = FALSE;
+    }
+    else
     if (Scr->ClientBorderWidth) {
     	tmp_win->frame_bw = tmp_win->old_bw;
     } else {
@@ -340,8 +389,13 @@ IconMgr *iconp;
     /*
      * do any prompting for position
      */
+    SetupOccupation (tmp_win);
+
     if (HandlingEvents && ask_user) {
-      if (Scr->RandomPlacement) {	/* just stick it somewhere */
+      if ((Scr->RandomPlacement == RP_ALL) ||
+          ((Scr->RandomPlacement == RP_UNMAPPED) &&
+	   ((tmp_win->wmhints && (tmp_win->wmhints->initial_state == IconicState)) ||
+	    (! OCCUPY (tmp_win, Scr->workSpaceMgr.activeWSPC))))) { /* just stick it somewhere */
 	if ((PlaceX + tmp_win->attr.width) > Scr->MyDisplayWidth)
 	    PlaceX = 50;
 	if ((PlaceY + tmp_win->attr.height) > Scr->MyDisplayHeight)
@@ -426,26 +480,30 @@ IconMgr *iconp;
 			      SIZE_VINDENT + Scr->SizeFont.font->ascent,
 			      tmp_win->name, namelen);
 
-	    AddingW = tmp_win->attr.width + bw2;
-	    AddingH = tmp_win->attr.height + tmp_win->title_height + bw2;
+	    AddingW = tmp_win->attr.width + bw2 + 2 * tmp_win->frame_bw3D;
+	    AddingH = tmp_win->attr.height + tmp_win->title_height +
+				bw2 + 2 * tmp_win->frame_bw3D;
 
 		MoveOutline(Scr->Root, AddingX, AddingY, AddingW, AddingH,
-			    tmp_win->frame_bw, tmp_win->title_height);
+			    tmp_win->frame_bw, tmp_win->title_height + tmp_win->frame_bw3D);
 
 	    XDrawImageString (dpy, Scr->SizeWindow, Scr->NormalGC, width,
 				SIZE_VINDENT + Scr->SizeFont.font->ascent, ": ", 2);
 	    DisplayPosition (tmp_win, AddingX, AddingY);
+	    /*SetFocus ((TwmWindow *) NULL, CurrentTime);*/
 	    while (TRUE)
 		{
 		XMaskEvent(dpy, ButtonPressMask | PointerMotionMask, &event);
 
-		if (Event.type == MotionNotify) {
+		if (event.type == MotionNotify) {
 		    /* discard any extra motion events before a release */
 		    while(XCheckMaskEvent(dpy,
-			ButtonMotionMask | ButtonPressMask, &Event))
-			if (Event.type == ButtonPress)
+			ButtonMotionMask | ButtonPressMask, &event))
+			if (event.type == ButtonPress)
 			    break;
 		}
+		event.xbutton.x_root -= Scr->MyDisplayX;
+		event.xbutton.y_root -= Scr->MyDisplayY;
 		
 		if (event.type == ButtonPress) {
 		  AddingX = event.xbutton.x_root;
@@ -459,14 +517,20 @@ IconMgr *iconp;
 		      AddingR = AddingX + AddingW;
 		      AddingB = AddingY + AddingH;
 		      
-		      if (AddingX < 0)
+		      if ((AddingX < 0) && ((Scr->MoveOffResistance < 0) 
+					    || (AddingX > -Scr->MoveOffResistance)))
 			AddingX = 0;
-		      if (AddingR > Scr->MyDisplayWidth)
+		      if ((AddingR > Scr->MyDisplayWidth) 
+			  && ((Scr->MoveOffResistance < 0) 
+			      || (AddingR < Scr->MyDisplayWidth + Scr->MoveOffResistance)))
 			AddingX = Scr->MyDisplayWidth - AddingW;
 		      
-		      if (AddingY < 0)
+		      if ((AddingY < 0) && ((Scr->MoveOffResistance < 0) 
+					    || (AddingY > -Scr->MoveOffResistance)))
 			AddingY = 0;
-		      if (AddingB > Scr->MyDisplayHeight)
+		      if ((AddingB > Scr->MyDisplayHeight)
+			  && ((Scr->MoveOffResistance < 0) 
+			      || (AddingB < Scr->MyDisplayHeight + Scr->MoveOffResistance)))
 			AddingY = Scr->MyDisplayHeight - AddingH;
 		      
  		    }
@@ -487,19 +551,24 @@ IconMgr *iconp;
 		    AddingR = AddingX + AddingW;
 		    AddingB = AddingY + AddingH;
 		    
-		    if (AddingX < 0)
+		    if ((AddingX < 0) && ((Scr->MoveOffResistance < 0) 
+					  || (AddingX > -Scr->MoveOffResistance)))
 		        AddingX = 0;
-		    if (AddingR > Scr->MyDisplayWidth)
+		    if ((AddingR > Scr->MyDisplayWidth) 
+			&& ((Scr->MoveOffResistance < 0) 
+			    || (AddingR < Scr->MyDisplayWidth + Scr->MoveOffResistance)))
 		        AddingX = Scr->MyDisplayWidth - AddingW;
 
-		    if (AddingY < 0)
+		    if ((AddingY < 0) && ((Scr->MoveOffResistance < 0) 
+					  || (AddingY > -Scr->MoveOffResistance)))
 			AddingY = 0;
-		    if (AddingB > Scr->MyDisplayHeight)
+		    if ((AddingB > Scr->MyDisplayHeight)
+			&& ((Scr->MoveOffResistance < 0) 
+			    || (AddingB < Scr->MyDisplayHeight + Scr->MoveOffResistance)))
 			AddingY = Scr->MyDisplayHeight - AddingH;
 		}
-
 		MoveOutline(Scr->Root, AddingX, AddingY, AddingW, AddingH,
-			    tmp_win->frame_bw, tmp_win->title_height);
+			    tmp_win->frame_bw, tmp_win->title_height + tmp_win->frame_bw3D);
 
 		DisplayPosition (tmp_win, AddingX, AddingY);
 	    }
@@ -543,13 +612,15 @@ IconMgr *iconp;
 		    XMaskEvent(dpy,
 			       ButtonReleaseMask | ButtonMotionMask, &event);
 
-		    if (Event.type == MotionNotify) {
+		    if (event.type == MotionNotify) {
 			/* discard any extra motion events before a release */
 			while(XCheckMaskEvent(dpy,
-			    ButtonMotionMask | ButtonReleaseMask, &Event))
-			    if (Event.type == ButtonRelease)
+			    ButtonMotionMask | ButtonReleaseMask, &event))
+			    if (event.type == ButtonRelease)
 				break;
 		    }
+		    event.xbutton.x_root -= Scr->MyDisplayX;
+		    event.xbutton.y_root -= Scr->MyDisplayY;
 
 		    if (event.type == ButtonRelease)
 		    {
@@ -583,7 +654,7 @@ IconMgr *iconp;
 	    } 
 	    else if (event.xbutton.button == Button3)
 	    {
-		int maxw = Scr->MyDisplayWidth - AddingX - bw2;
+		int maxw = Scr->MyDisplayWidth  - AddingX - bw2;
 		int maxh = Scr->MyDisplayHeight - AddingY - bw2;
 
 		/*
@@ -597,6 +668,7 @@ IconMgr *iconp;
 		ConstrainSize (tmp_win, &AddingW, &AddingH);  /* w/o borders */
 		AddingW += bw2;
 		AddingH += bw2;
+		XMaskEvent(dpy, ButtonReleaseMask, &event);
 	    }
 	    else
 	    {
@@ -610,8 +682,9 @@ IconMgr *iconp;
 
 	    tmp_win->attr.x = AddingX;
 	    tmp_win->attr.y = AddingY + tmp_win->title_height;
-	    tmp_win->attr.width = AddingW - bw2;
-	    tmp_win->attr.height = AddingH - tmp_win->title_height - bw2;
+	    tmp_win->attr.width = AddingW - bw2 - 2 * tmp_win->frame_bw3D;
+	    tmp_win->attr.height = AddingH - tmp_win->title_height -
+				bw2 - 2 * tmp_win->frame_bw3D;
 
 	    XUngrabServer(dpy);
 	}
@@ -630,15 +703,13 @@ IconMgr *iconp;
 #endif
 
     if (!Scr->ClientBorderWidth) {	/* need to adjust for twm borders */
-	int delta = tmp_win->attr.border_width - tmp_win->frame_bw;
+	int delta = tmp_win->attr.border_width - tmp_win->frame_bw - tmp_win->frame_bw3D;
 	tmp_win->attr.x += gravx * delta;
 	tmp_win->attr.y += gravy * delta;
     }
 
     tmp_win->title_width = tmp_win->attr.width;
 
-    if (tmp_win->old_bw) XSetWindowBorderWidth (dpy, tmp_win->w, 0);
-
     tmp_win->name_width = XTextWidth(Scr->TitleBarFont.font, tmp_win->name,
 				     namelen);
 
@@ -650,6 +721,8 @@ IconMgr *iconp;
     if (tmp_win->icon_name == NULL)
 	tmp_win->icon_name = tmp_win->name;
 
+    if (tmp_win->old_bw) XSetWindowBorderWidth (dpy, tmp_win->w, 0);
+
     tmp_win->iconified = FALSE;
     tmp_win->isicon = FALSE;
     tmp_win->icon_on = FALSE;
@@ -680,14 +753,17 @@ IconMgr *iconp;
 
     /* get all the colors for the window */
 
-    tmp_win->border = Scr->BorderColor;
+    tmp_win->borderC.fore     = Scr->BorderColorC.fore;
+    tmp_win->borderC.back     = Scr->BorderColorC.back;
     tmp_win->border_tile.fore = Scr->BorderTileC.fore;
     tmp_win->border_tile.back = Scr->BorderTileC.back;
-    tmp_win->title.fore = Scr->TitleC.fore;
-    tmp_win->title.back = Scr->TitleC.back;
+    tmp_win->title.fore       = Scr->TitleC.fore;
+    tmp_win->title.back       = Scr->TitleC.back;
 
     GetColorFromList(Scr->BorderColorL, tmp_win->full_name, &tmp_win->class,
-	&tmp_win->border);
+	&tmp_win->borderC.fore);
+    GetColorFromList(Scr->BorderColorL, tmp_win->full_name, &tmp_win->class,
+	&tmp_win->borderC.back);
     GetColorFromList(Scr->BorderTileForegroundL, tmp_win->full_name,
 	&tmp_win->class, &tmp_win->border_tile.fore);
     GetColorFromList(Scr->BorderTileBackgroundL, tmp_win->full_name,
@@ -697,27 +773,39 @@ IconMgr *iconp;
     GetColorFromList(Scr->TitleBackgroundL, tmp_win->full_name, &tmp_win->class,
 	&tmp_win->title.back);
 
-    if (Scr->use3Dtitles && !Scr->BeNiceToColormap) GetShadeColors (&tmp_win->title);
-
+    if (Scr->use3Dtitles  && !Scr->BeNiceToColormap) GetShadeColors (&tmp_win->title);
+    if (Scr->use3Dborders && !Scr->BeNiceToColormap) {
+	GetShadeColors (&tmp_win->borderC);
+	GetShadeColors (&tmp_win->border_tile);
+    }
     /* create windows */
 
-    tmp_win->frame_x = tmp_win->attr.x + tmp_win->old_bw - tmp_win->frame_bw;
+    tmp_win->frame_x = tmp_win->attr.x + tmp_win->old_bw - tmp_win->frame_bw
+			- tmp_win->frame_bw3D;
     tmp_win->frame_y = tmp_win->attr.y - tmp_win->title_height +
-	tmp_win->old_bw - tmp_win->frame_bw;
-    tmp_win->frame_width = tmp_win->attr.width;
-    tmp_win->frame_height = tmp_win->attr.height + tmp_win->title_height;
+	tmp_win->old_bw - tmp_win->frame_bw - tmp_win->frame_bw3D;
+    tmp_win->frame_width = tmp_win->attr.width + 2 * tmp_win->frame_bw3D;
+    tmp_win->frame_height = tmp_win->attr.height + tmp_win->title_height +
+				2 * tmp_win->frame_bw3D;
+
+    ConstrainSize (tmp_win, &tmp_win->frame_width, &tmp_win->frame_height);
 
     valuemask = CWBackPixmap | CWBorderPixel | CWCursor | CWEventMask;
     attributes.background_pixmap = None;
-    attributes.border_pixel = tmp_win->border;
+    attributes.border_pixel = tmp_win->borderC.back;
     attributes.cursor = Scr->FrameCursor;
     attributes.event_mask = (SubstructureRedirectMask | 
 			     ButtonPressMask | ButtonReleaseMask |
-			     EnterWindowMask | LeaveWindowMask);
+			     EnterWindowMask | LeaveWindowMask | ExposureMask);
+    if (Scr->BorderCursors) attributes.event_mask |= PointerMotionMask;
     if (tmp_win->attr.save_under) {
 	attributes.save_under = True;
 	valuemask |= CWSaveUnder;
     }
+    if (tmp_win->hints.flags & PWinGravity) {
+	attributes.win_gravity = tmp_win->hints.win_gravity;
+	valuemask |= CWWinGravity;
+    }
 
     tmp_win->frame = XCreateWindow (dpy, Scr->Root, tmp_win->frame_x,
 				    tmp_win->frame_y, 
@@ -730,14 +818,15 @@ IconMgr *iconp;
     
     if (tmp_win->title_height)
     {
-	valuemask = (CWEventMask | CWBorderPixel | CWBackPixel);
+	valuemask = (CWEventMask | CWDontPropagate | CWBorderPixel | CWBackPixel);
 	attributes.event_mask = (KeyPressMask | ButtonPressMask |
 				 ButtonReleaseMask | ExposureMask);
-	attributes.border_pixel = tmp_win->border;
+	attributes.do_not_propagate_mask = PointerMotionMask;
+	attributes.border_pixel = tmp_win->borderC.back;
 	attributes.background_pixel = tmp_win->title.back;
 	tmp_win->title_w = XCreateWindow (dpy, tmp_win->frame, 
-					  -tmp_win->frame_bw,
-					  -tmp_win->frame_bw,
+					  tmp_win->frame_bw3D - tmp_win->frame_bw,
+					  tmp_win->frame_bw3D - tmp_win->frame_bw,
 					  (unsigned int) tmp_win->attr.width, 
 					  (unsigned int) Scr->TitleHeight,
 					  (unsigned int) tmp_win->frame_bw,
@@ -755,34 +844,41 @@ IconMgr *iconp;
     {
 	if (Scr->use3Dtitles && (Scr->Monochrome != COLOR))
 	    tmp_win->gray = XCreatePixmapFromBitmapData(dpy, Scr->Root, 
-		black_bits, gray_width, gray_height, 
+		(char*)black_bits, gray_width, gray_height, 
 		tmp_win->border_tile.fore, tmp_win->border_tile.back,
 		Scr->d_depth);
 	else
 	    tmp_win->gray = XCreatePixmapFromBitmapData(dpy, Scr->Root, 
-		gray_bits, gray_width, gray_height, 
+		(char*)gray_bits, gray_width, gray_height, 
 		tmp_win->border_tile.fore, tmp_win->border_tile.back,
 		Scr->d_depth);
 
-	SetBorder (tmp_win, False);
+	tmp_win->hasfocusvisible = True;
+	SetFocusVisualAttributes (tmp_win, False);
     }
     else
 	tmp_win->gray = None;
 
+    RaiseWindow(tmp_win);
 	
     if (tmp_win->title_w) {
-	CreateWindowTitlebarButtons (tmp_win);
 	ComputeTitleLocation (tmp_win);
+	CreateWindowTitlebarButtons (tmp_win);
 	XMoveWindow (dpy, tmp_win->title_w,
 		     tmp_win->title_x, tmp_win->title_y);
 	XDefineCursor(dpy, tmp_win->title_w, Scr->TitleCursor);
     }
+    else {
+	tmp_win->title_x = tmp_win->frame_bw3D - tmp_win->frame_bw;
+	tmp_win->title_y = tmp_win->frame_bw3D - tmp_win->frame_bw;
+    }
 
     valuemask = (CWEventMask | CWDontPropagate);
     attributes.event_mask = (StructureNotifyMask | PropertyChangeMask |
 			     ColormapChangeMask | VisibilityChangeMask |
+			     FocusChangeMask | 
 			     EnterWindowMask | LeaveWindowMask);
-    attributes.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
+    attributes.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
     XChangeWindowAttributes (dpy, tmp_win->w, valuemask, &attributes);
 
     if (HasShape)
@@ -807,7 +903,8 @@ IconMgr *iconp;
     if (!tmp_win->iconmgr)
 	XAddToSaveSet(dpy, tmp_win->w);
 	
-    XReparentWindow(dpy, tmp_win->w, tmp_win->frame, 0, tmp_win->title_height);
+    XReparentWindow(dpy, tmp_win->w, tmp_win->frame, tmp_win->frame_bw3D,
+		tmp_win->title_height + tmp_win->frame_bw3D);
     /*
      * Reparenting generates an UnmapNotify event, followed by a MapNotify.
      * Set the map state to FALSE to prevent a transition back to
@@ -829,15 +926,19 @@ IconMgr *iconp;
     {
 	GrabButtons(tmp_win);
 	GrabKeys(tmp_win);
+	if (Scr->ClickToFocus &&
+	    tmp_win->wmhints && tmp_win->wmhints->input &&
+	    (w != Scr->workSpaceMgr.workspaceWindow.w) &&
+	    (w != Scr->workSpaceMgr.occupyWindow.w)) ClickToFocusGrab (tmp_win);
     }
 
-    SetupOccupation (tmp_win);
     (void) AddIconManager(tmp_win);
 
     XSaveContext(dpy, tmp_win->w, TwmContext, (caddr_t) tmp_win);
     XSaveContext(dpy, tmp_win->w, ScreenContext, (caddr_t) Scr);
     XSaveContext(dpy, tmp_win->frame, TwmContext, (caddr_t) tmp_win);
     XSaveContext(dpy, tmp_win->frame, ScreenContext, (caddr_t) Scr);
+
     if (tmp_win->title_height)
     {
 	int i;
@@ -851,10 +952,15 @@ IconMgr *iconp;
 	    XSaveContext(dpy, tmp_win->titlebuttons[i].window, ScreenContext,
 			 (caddr_t) Scr);
 	}
-	if (tmp_win->hilite_w)
+	if (tmp_win->hilite_wl)
 	{
-	    XSaveContext(dpy, tmp_win->hilite_w, TwmContext, (caddr_t)tmp_win);
-	    XSaveContext(dpy, tmp_win->hilite_w, ScreenContext, (caddr_t)Scr);
+	    XSaveContext(dpy, tmp_win->hilite_wl, TwmContext, (caddr_t)tmp_win);
+	    XSaveContext(dpy, tmp_win->hilite_wl, ScreenContext, (caddr_t)Scr);
+	}
+	if (tmp_win->hilite_wr)
+	{
+	    XSaveContext(dpy, tmp_win->hilite_wr, TwmContext, (caddr_t)tmp_win);
+	    XSaveContext(dpy, tmp_win->hilite_wr, ScreenContext, (caddr_t)Scr);
 	}
     }
 
@@ -907,20 +1013,11 @@ static void do_add_binding (button, context, modifier, func)
     int button, context, modifier;
     int func;
 { 
-    if (Scr->Mouse[button][context][modifier]) return;	/* already defined */
-
-    Scr->Mouse[button][context][modifier] = (MouseButton*) malloc (sizeof (MouseButton));
-    Scr->Mouse[button][context][modifier]->func = func;
-    Scr->Mouse[button][context][modifier]->item = NULL;
+    AddFuncButton (button, context, modifier, func, NULL, NULL);
 }
 
 AddDefaultBindings ()
 {
-    /*
-     * The bindings are stored in Scr->Mouse, indexed by
-     * Mouse[button_number][C_context][modifier].
-     */
-
 #define NoModifierMask 0
 
     do_add_binding (Button1, C_TITLE, NoModifierMask, F_MOVE);
@@ -948,28 +1045,40 @@ AddDefaultBindings ()
  ***********************************************************************
  */
 
+#define AltMask (Alt1Mask | Alt2Mask | Alt3Mask | Alt4Mask | Alt5Mask)
+
 void
 GrabButtons(tmp_win)
 TwmWindow *tmp_win;
 {
-    int i, j;
+    FuncButton *tmp;
 
-    for (i = 0; i < MAX_BUTTONS+1; i++)
-    {
-	for (j = 0; j < MOD_SIZE; j++)
-	{
-	    if ((Scr->Mouse[i][C_WINDOW][j] != (MouseButton*) 0) &&
-	        (Scr->Mouse[i][C_WINDOW][j]->func != 0))
-	    {
-	        /* twm used to do this grab on the application main window,
-                 * tmp_win->w . This was not ICCCM complient and was changed.
-		 */
-		XGrabButton(dpy, i, j, tmp_win->frame, 
-			    True, ButtonPressMask | ButtonReleaseMask,
-			    GrabModeAsync, GrabModeAsync, None, 
-			    Scr->FrameCursor);
-	    }
-	}
+    for (tmp = Scr->FuncButtonRoot.next; tmp != NULL; tmp = tmp->next) {
+	if ((tmp->cont != C_WINDOW) || (tmp->func == 0)) continue;
+	XGrabButton (dpy, tmp->num, tmp->mods, tmp_win->frame, 
+		    True, ButtonPressMask | ButtonReleaseMask,
+		    GrabModeAsync, GrabModeAsync, None, 
+		    Scr->FrameCursor);
+	    
+    }
+}
+
+void ClickToFocusGrab (tmp_win)
+TwmWindow *tmp_win;
+{
+    if (tmp_win) {
+	XGrabButton (dpy, AnyButton, None, tmp_win->w, 
+		    True, ButtonPressMask | ButtonReleaseMask,
+		    GrabModeAsync, GrabModeAsync, None, 
+		    Scr->FrameCursor);
+    }
+}
+
+void ClickToFocusUngrab (tmp_win)
+TwmWindow *tmp_win;
+{
+    if (tmp_win) {
+	XUngrabButton (dpy, AnyButton, AnyModifier, tmp_win->w);
     }
 }
 
@@ -998,6 +1107,7 @@ TwmWindow *tmp_win;
 	switch (tmp->cont)
 	{
 	case C_WINDOW:
+	    if (tmp->mods & AltMask) break;
 	    XGrabKey(dpy, tmp->keycode, tmp->mods, tmp_win->w, True,
 		GrabModeAsync, GrabModeAsync);
 	    break;
@@ -1044,18 +1154,21 @@ TwmWindow *tmp_win;
     }
 }
 
-static Window CreateHighlightWindow (tmp_win)
+static void CreateHighlightWindows (tmp_win)
     TwmWindow *tmp_win;
 {
     XSetWindowAttributes attributes;	/* attributes for create windows */
-    Pixmap pm = None;
     GC gc;
     XGCValues gcv;
     unsigned long valuemask;
     int h = (Scr->TitleHeight - 2 * Scr->FramePadding);
-    Window w;
-
+    int y = Scr->FramePadding;
 
+    if (! tmp_win->titlehighlight) {
+	tmp_win->hilite_wl = (Window) 0;
+	tmp_win->hilite_wr = (Window) 0;
+	return;
+    }
     /*
      * If a special highlight pixmap was given, use that.  Otherwise,
      * use a nice, even gray pattern.  The old horizontal lines look really
@@ -1068,65 +1181,70 @@ static Window CreateHighlightWindow (tmp_win)
      * file.  If all else fails, use the foreground color to look like a 
      * solid line.
      */
-    if (!Scr->hilitePm) {
-	if (Scr->use3Dtitles && (Scr->Monochrome != COLOR))
-	    Scr->hilitePm = XCreateBitmapFromData (dpy, tmp_win->title_w, 
-					       black_bits, gray_width, 
-					       gray_height);
-	else
-	    Scr->hilitePm = XCreateBitmapFromData (dpy, tmp_win->title_w, 
-					       gray_bits, gray_width, 
-					       gray_height);
-	Scr->hilite_pm_width = gray_width;
-	Scr->hilite_pm_height = gray_height;
+
+    if (! tmp_win->HiliteImage) {
+	if (Scr->HighlightPixmapName) {
+	    tmp_win->HiliteImage = GetImage (Scr->HighlightPixmapName, tmp_win->title);
+	}
     }
-    if (Scr->hilitePm) {
-      if (Scr->hilite_pm_depth == Scr->d_depth) {
-	pm = Scr->hilitePm;
-      }
-      else {
-	pm = XCreatePixmap (dpy, tmp_win->title_w,
-			    Scr->hilite_pm_width, Scr->hilite_pm_height,
-			    Scr->d_depth);
+    if (! tmp_win->HiliteImage) {
+	Pixmap pm = None;
+	Pixmap bm = None;
+
+	if (Scr->use3Dtitles && (Scr->Monochrome != COLOR))
+	    bm = XCreateBitmapFromData (dpy, tmp_win->title_w, 
+					(char*)black_bits, gray_width, gray_height);
+	else
+	    bm = XCreateBitmapFromData (dpy, tmp_win->title_w, 
+					(char*)gray_bits, gray_width, gray_height);
+
+	pm = XCreatePixmap (dpy, tmp_win->title_w, gray_width, gray_height, Scr->d_depth);
 	gcv.foreground = tmp_win->title.fore;
 	gcv.background = tmp_win->title.back;
 	gcv.graphics_exposures = False;
-	gc = XCreateGC (dpy, pm,
-			(GCForeground|GCBackground|GCGraphicsExposures),
-			&gcv);
+	gc = XCreateGC (dpy, pm, (GCForeground|GCBackground|GCGraphicsExposures), &gcv);
 	if (gc) {
-	    XCopyPlane (dpy, Scr->hilitePm, pm, gc, 0, 0, 
-			Scr->hilite_pm_width, Scr->hilite_pm_height,
-			0, 0, 1);
+	    XCopyPlane (dpy, bm, pm, gc, 0, 0, gray_width, gray_height, 0, 0, 1);
+	    tmp_win->HiliteImage = (Image*) malloc (sizeof (struct _Image));
+	    tmp_win->HiliteImage->pixmap = pm;
+	    tmp_win->HiliteImage->width  = gray_width;
+	    tmp_win->HiliteImage->height = gray_height;
+	    tmp_win->HiliteImage->mask   = None;
+	    tmp_win->HiliteImage->next   = None;
 	    XFreeGC (dpy, gc);
 	} else {
 	    XFreePixmap (dpy, pm);
 	    pm = None;
 	}
-      }
+	XFreePixmap (dpy, bm);
     }
-    if (pm) {
+    if (tmp_win->HiliteImage) {
 	valuemask = CWBackPixmap;
-	attributes.background_pixmap = pm;
+	attributes.background_pixmap = tmp_win->HiliteImage->pixmap;
     } else {
 	valuemask = CWBackPixel;
 	attributes.background_pixel = tmp_win->title.fore;
     }
 
-    if (Scr->use3Dtitles)
-	w = XCreateWindow (dpy, tmp_win->title_w, 0, Scr->FramePadding + 2,
-		       (unsigned int) Scr->TBInfo.width, (unsigned int) (h - 4),
-		       (unsigned int) 0,
-		       Scr->d_depth, (unsigned int) CopyFromParent,
-		       Scr->d_visual, valuemask, &attributes);
+    if (Scr->use3Dtitles) {
+	y += Scr->TitleShadowDepth;
+	h -= 2 * Scr->TitleShadowDepth;
+    }
+    if (Scr->TitleJustification == J_LEFT)
+	tmp_win->hilite_wl = (Window) 0;
     else
-	w = XCreateWindow (dpy, tmp_win->title_w, 0, Scr->FramePadding,
+	tmp_win->hilite_wl = XCreateWindow (dpy, tmp_win->title_w, 0, y,
 		       (unsigned int) Scr->TBInfo.width, (unsigned int) h,
-		       (unsigned int) 0,
-		       Scr->d_depth, (unsigned int) CopyFromParent,
+		       (unsigned int) 0, Scr->d_depth, (unsigned int) CopyFromParent,
+		       Scr->d_visual, valuemask, &attributes);
+
+    if (Scr->TitleJustification == J_RIGHT)
+	tmp_win->hilite_wr = (Window) 0;
+    else
+	tmp_win->hilite_wr = XCreateWindow (dpy, tmp_win->title_w, 0, y,
+		       (unsigned int) Scr->TBInfo.width, (unsigned int) h,
+		       (unsigned int) 0,  Scr->d_depth, (unsigned int) CopyFromParent,
 		       Scr->d_visual, valuemask, &attributes);
-    if (pm && (Scr->hilite_pm_depth != Scr->d_depth)) XFreePixmap (dpy, pm);
-    return w;
 }
 
 
@@ -1135,16 +1253,14 @@ void ComputeCommonTitleOffsets ()
     int buttonwidth = (Scr->TBInfo.width + Scr->TBInfo.pad);
 
     Scr->TBInfo.leftx = Scr->TBInfo.rightoff = Scr->FramePadding;
-    if (Scr->TBInfo.nleft > 0)
-      Scr->TBInfo.leftx += Scr->ButtonIndent;
-    Scr->TBInfo.titlex = (Scr->TBInfo.leftx +
-			  (Scr->TBInfo.nleft * buttonwidth) - Scr->TBInfo.pad +
-			  Scr->TitlePadding);
-    if (Scr->TBInfo.nright > 0)
-      Scr->TBInfo.rightoff += (Scr->ButtonIndent +
+    if (Scr->TBInfo.nleft  > 0) Scr->TBInfo.leftx    += Scr->ButtonIndent;
+    if (Scr->TBInfo.nright > 0) Scr->TBInfo.rightoff += (Scr->ButtonIndent +
 			       ((Scr->TBInfo.nright * buttonwidth) -
 				Scr->TBInfo.pad));
 
+    Scr->TBInfo.titlex = (Scr->TBInfo.leftx +
+			  (Scr->TBInfo.nleft * buttonwidth) - Scr->TBInfo.pad +
+			  Scr->TitlePadding);
     return;
 }
 
@@ -1152,14 +1268,41 @@ void ComputeWindowTitleOffsets (tmp_win, width, squeeze)
     TwmWindow *tmp_win;
     Bool squeeze;
 {
-    tmp_win->highlightx = (Scr->TBInfo.titlex + tmp_win->name_width);
-    if (Scr->use3Dtitles) tmp_win->highlightx += 6;
-    if (tmp_win->hilite_w || Scr->TBInfo.nright > 0) 
-      tmp_win->highlightx += Scr->TitlePadding;
+    int titlew = width - Scr->TBInfo.titlex - Scr->TBInfo.rightoff;
+
+    switch (Scr->TitleJustification) {
+	case J_LEFT :
+	    tmp_win->name_x = Scr->TBInfo.titlex;
+	    if (Scr->use3Dtitles) tmp_win->name_x += Scr->TitleShadowDepth + 2;
+	    break;
+	case J_CENTER :
+	    tmp_win->name_x = Scr->TBInfo.titlex + (titlew - tmp_win->name_width) / 2;
+	    break;
+	case J_RIGHT :
+	    tmp_win->name_x = Scr->TBInfo.titlex + titlew - tmp_win->name_width;
+	    if (Scr->use3Dtitles) tmp_win->name_x -= Scr->TitleShadowDepth - 2;
+	    break;
+    }
+    if (Scr->use3Dtitles) {
+	if (tmp_win->name_x < (Scr->TBInfo.titlex + 2 * Scr->TitleShadowDepth))
+		tmp_win->name_x = Scr->TBInfo.titlex + 2 * Scr->TitleShadowDepth;
+    }
+    else
+    if (tmp_win->name_x < Scr->TBInfo.titlex) {
+	tmp_win->name_x = Scr->TBInfo.titlex;
+    }
+    tmp_win->highlightxl = Scr->TBInfo.titlex;
+    tmp_win->highlightxr = tmp_win->name_x + tmp_win->name_width + 2;
+
+    if (Scr->use3Dtitles) {
+	tmp_win->highlightxl += Scr->TitleShadowDepth;
+    }
+    if (tmp_win->hilite_wr || Scr->TBInfo.nright > 0) 
+      tmp_win->highlightxr += Scr->TitlePadding;
     tmp_win->rightx = width - Scr->TBInfo.rightoff;
     if (squeeze && tmp_win->squeeze_info) {
-	int rx = (tmp_win->highlightx + 
-		  (tmp_win->hilite_w
+	int rx = (tmp_win->highlightxr + 
+		  (tmp_win->hilite_wr
 		    ? Scr->TBInfo.width * 2 : 0) +
 		  (Scr->TBInfo.nright > 0 ? Scr->TitlePadding : 0) +
 		  Scr->FramePadding);
@@ -1177,14 +1320,14 @@ void ComputeWindowTitleOffsets (tmp_win, width, squeeze)
 void ComputeTitleLocation (tmp)
     register TwmWindow *tmp;
 {
-    tmp->title_x = -tmp->frame_bw;
-    tmp->title_y = -tmp->frame_bw;
+    tmp->title_x = tmp->frame_bw3D - tmp->frame_bw;
+    tmp->title_y = tmp->frame_bw3D - tmp->frame_bw;
 
     if (tmp->squeeze_info) {
 	register SqueezeInfo *si = tmp->squeeze_info;
 	int basex;
 	int maxwidth = tmp->frame_width;
-	int tw = tmp->title_width;
+	int tw = tmp->title_width + 2 * tmp->frame_bw3D;
 
 	/*
 	 * figure label base from squeeze info (justification fraction)
@@ -1216,11 +1359,11 @@ void ComputeTitleLocation (tmp)
 	    basex -= tw - 1;
 	    break;
 	}
-	if (basex > maxwidth - tw + 1)
-	  basex = maxwidth - tw + 1;
+	if (basex > maxwidth - tw)
+	  basex = maxwidth - tw;
 	if (basex < 0) basex = 0;
 
-	tmp->title_x = basex - tmp->frame_bw;
+	tmp->title_x = basex - tmp->frame_bw + tmp->frame_bw3D;
     }
 }
 
@@ -1236,7 +1379,8 @@ static void CreateWindowTitlebarButtons (tmp_win)
 
     if (tmp_win->title_height == 0)
     {
-	tmp_win->hilite_w = 0;
+	tmp_win->hilite_wl = 0;
+	tmp_win->hilite_wr = 0;
 	return;
     }
 
@@ -1288,26 +1432,23 @@ static void CreateWindowTitlebarButtons (tmp_win)
 					     0, (unsigned int) CopyFromParent,
 					     (Visual *) CopyFromParent,
 					     valuemask, &attributes);
-		if (Scr->use3Dtitles && (tb->depth == Scr->d_depth) &&
-		    (tmp_win->title.back != Scr->TitleC.back)) {
-		    unsigned int junkw, junkh, junkd;
-
-		    tbw->bitmap = FindPixmap (tb->name, &junkw, &junkh, &junkd, tmp_win->title);
-		}
-		else {
-		    tbw->bitmap = tb->bitmap;
+		tbw->image = GetImage (tb->name, tmp_win->title);
+		if (! tbw->image) {
+		    tbw->image = GetImage (TBPM_QUESTION, tmp_win->title);
+		    if (! tbw->image) {		/* cannot happen (see util.c) */
+			fprintf (stderr, "%s:  unable to add titlebar button \"%s\"\n",
+				 ProgramName, tb->name);
+		    }
 		}
 		tbw->info = tb;
 	    }
 	}
     }
 
-    tmp_win->hilite_w = (tmp_win->titlehighlight 
-			 ? CreateHighlightWindow (tmp_win) : None);
-
+    CreateHighlightWindows (tmp_win);
     XMapSubwindows(dpy, tmp_win->title_w);
-    if (tmp_win->hilite_w)
-      XUnmapWindow(dpy, tmp_win->hilite_w);
+    if (tmp_win->hilite_wl) XUnmapWindow(dpy, tmp_win->hilite_wl);
+    if (tmp_win->hilite_wr) XUnmapWindow(dpy, tmp_win->hilite_wr);
     return;
 }
 
@@ -1315,36 +1456,13 @@ static void CreateWindowTitlebarButtons (tmp_win)
 SetHighlightPixmap (filename)
     char *filename;
 {
-    Pixmap pm;
-
-#ifdef XPM
-    if ((filename [0] == '@') || (strncmp (filename, "xpm:", 4) == 0)) {
-	XpmIcon *xpmicon;
-	int startn;
-
-	if (filename [0] == '@') startn = 1; else startn = 4;
-	xpmicon = GetXpmPixmap (&filename [startn]);
-	if (xpmicon != None) {
-	    if (Scr->hilitePm) XFreePixmap (dpy, Scr->hilitePm);
-	    Scr->hilitePm         = xpmicon->pixmap;
-	    Scr->hilite_pm_depth  = Scr->d_depth;
-	    Scr->hilite_pm_width  = xpmicon->attributes.width;
-	    Scr->hilite_pm_height = xpmicon->attributes.height;
-	}
-	return;
-    }
+#ifdef VMS
+    char *ftemp;
+    ftemp = (char *) malloc((strlen(filename)+1)*sizeof(char));
+    Scr->HighlightPixmapName = strcpy (ftemp,filename);
+#else
+    Scr->HighlightPixmapName = (char*) strdup (filename);
 #endif
-    pm = GetBitmap (filename);
-
-    if (pm) {
-	if (Scr->hilitePm) {
-	    XFreePixmap (dpy, Scr->hilitePm);
-	}
-	Scr->hilitePm = pm;
-	Scr->hilite_pm_depth  = 1;
-	Scr->hilite_pm_width  = JunkWidth;
-	Scr->hilite_pm_height = JunkHeight;
-    }
 }
 
 
@@ -1571,7 +1689,7 @@ FetchWmColormapWindows (tmp)
 	  XFree ((char *) cmap_windows);
     }
 
-    return;
+    return (0);
 }
 
 
@@ -1594,8 +1712,42 @@ void GetWindowSizeHints (tmp)
 	int right =  tmp->attr.x + tmp->attr.width + 2 * tmp->old_bw;
 	int bottom = tmp->attr.y + tmp->attr.height + 2 * tmp->old_bw;
 	tmp->hints.win_gravity = 
-	  gravs[((Scr->MyDisplayHeight - bottom < tmp->title_height) ? 0 : 2) |
-		((Scr->MyDisplayWidth - right   < tmp->title_height) ? 0 : 1)];
+	  gravs[((Scr->MyDisplayHeight - bottom <
+		tmp->title_height + 2 * tmp->frame_bw3D) ? 0 : 2) |
+		((Scr->MyDisplayWidth - right   <
+		tmp->title_height + 2 * tmp->frame_bw3D) ? 0 : 1)];
 	tmp->hints.flags |= PWinGravity;
     }
 }
+
+AnimateButton (tbw)
+TBWindow *tbw;
+{
+    Image	*image;
+    XSetWindowAttributes attr;
+
+    image = tbw->image;
+    attr.background_pixmap = image->pixmap;
+    XChangeWindowAttributes (dpy, tbw->window, CWBackPixmap, &attr);
+    XClearWindow (dpy, tbw->window);
+    tbw->image = image->next;
+}
+
+AnimateHighlight (t)
+TwmWindow *t;
+{
+    Image	*image;
+    XSetWindowAttributes attr;
+
+    image = t->HiliteImage;
+    attr.background_pixmap = image->pixmap;
+    if (t->hilite_wl) {
+	XChangeWindowAttributes (dpy, t->hilite_wl, CWBackPixmap, &attr);
+	XClearWindow (dpy, t->hilite_wl);
+    }
+    if (t->hilite_wr) {
+	XChangeWindowAttributes (dpy, t->hilite_wr, CWBackPixmap, &attr);
+	XClearWindow (dpy, t->hilite_wr);
+    }
+    t->HiliteImage = image->next;
+}
