@@ -24,6 +24,31 @@
 /**    TORTIOUS ACTION, ARISING OUT OF OR IN  CONNECTION  WITH  THE  USE    **/
 /**    OR PERFORMANCE OF THIS SOFTWARE.                                     **/
 /*****************************************************************************/
+/* 
+ *  [ ctwm ]
+ *
+ *  Copyright 1992 Claude Lecommandeur.
+ *            
+ * Permission to use, copy, modify  and distribute this software  [ctwm] and
+ * its documentation for any purpose is hereby granted without fee, provided
+ * that the above  copyright notice appear  in all copies and that both that
+ * copyright notice and this permission notice appear in supporting documen-
+ * tation, and that the name of  Claude Lecommandeur not be used in adverti-
+ * sing or  publicity  pertaining to  distribution of  the software  without
+ * specific, written prior permission. Claude Lecommandeur make no represen-
+ * tations  about the suitability  of this software  for any purpose.  It is
+ * provided "as is" without express or implied warranty.
+ *
+ * Claude Lecommandeur DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * INCLUDING ALL  IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS.  IN NO
+ * EVENT SHALL  Claude Lecommandeur  BE LIABLE FOR ANY SPECIAL,  INDIRECT OR
+ * CONSEQUENTIAL  DAMAGES OR ANY  DAMAGES WHATSOEVER  RESULTING FROM LOSS OF
+ * USE, DATA  OR PROFITS,  WHETHER IN AN ACTION  OF CONTRACT,  NEGLIGENCE OR
+ * OTHER  TORTIOUS ACTION,  ARISING OUT OF OR IN  CONNECTION WITH THE USE OR
+ * PERFORMANCE OF THIS SOFTWARE.
+ *
+ * Author:  Claude Lecommandeur [ lecom@sic.epfl.ch ][ April 1992 ]
+ */
 
 
 /***********************************************************************
@@ -43,15 +68,28 @@
  ***********************************************************************/
 
 #include <stdio.h>
+#if defined(sony_news)
+#  include <ctype.h>
+#endif
+#ifdef VMS
+#include <ctype.h>
+#include <decw$include/Xos.h>
+#include <X11Xmu/CharSet.h>
+#else
 #include <X11/Xos.h>
 #include <X11/Xmu/CharSet.h>
+#endif
 #include "twm.h"
 #include "screen.h"
 #include "menus.h"
 #include "util.h"
 #include "gram.h"
 #include "parse.h"
+#ifdef VMS
+#include <decw$include/Xatom.h> 
+#else
 #include <X11/Xatom.h> 
+#endif
 
 /* For m4... */
 #ifdef USEM4
@@ -62,10 +100,19 @@
 #endif
 
 #ifndef SYSTEM_INIT_FILE
+#ifdef VMS
+#define SYSTEM_INIT_FILE "DECW$SYSTEM_DEFAULTS:SYSTEM.CTWMRC"
+#else
 #define SYSTEM_INIT_FILE "/usr/lib/X11/twm/system.twmrc"
 #endif
+#endif
 #define BUF_LEN 300
 
+static int ParseJustification ();
+static int ParseAlignement ();
+static int ParseRandomPlacement ();
+static int ParseButtonStyle ();
+
 static FILE *twmrc;
 static int ptr = 0;
 static int len = 0;
@@ -95,6 +142,9 @@ int (*twmInputFunc)();
 
 extern char *defTwmrc[];		/* default bindings */
 
+#ifdef SOUNDS
+extern int set_sound_host();
+#endif
 
 /***********************************************************************
  *
@@ -147,20 +197,81 @@ int ParseTwmrc (filename)
 
     /*
      * Check for the twmrc file in the following order:
-     *   0.  -f filename
-     *   1.  .ctwmrc.#
-     *   2.  .ctwmrc
-     *   3.  .twmrc.#
-     *   4.  .twmrc
-     *   5.  system.ctwmrc
+     *       Unix                  |   VMS
+     *   0.  -f filename.#         | -f filename_#
+     *   1.  -f filename           | -f filename
+     *   2.  .ctwmrc.#             | ctwm.rc_#
+     *   3.  .ctwmrc               | ctwm.rc
+     *   4.  .twmrc.#              | twm.rc_#
+     *   5.  .twmrc                | twm.rc
+     *   6.  system.ctwmrc         | system.ctwmrc
      */
-    for (twmrc = NULL, i = 0; !twmrc && i < 6; i++) {
+    for (twmrc = NULL, i = 0; !twmrc && i < 7; i++) {
 	switch (i) {
-	  case 0:			/* -f filename */
+#ifdef VMS
+	  case 0:
+	    if (filename != NULL)  {
+	       cp = tmpfilename;
+	       (void) sprintf (tmpfilename, "%s_%d", filename, Scr->screen);
+	    } else
+	       cp = filename;
+	    break;
+
+	  case 1:
 	    cp = filename;
 	    break;
 
-	  case 1:			/* ~/.ctwmrc.screennum */
+	  case 2:
+	    if (!filename) {
+		home = getenv ("DECW$USER_DEFAULTS");
+		if (home) {
+		    homelen = strlen (home);
+		    cp = tmpfilename;
+		    (void) sprintf (tmpfilename, "%sctwm.rc_%d",
+				    home, Scr->screen);
+		    break;
+		}
+	    }
+	    continue;
+
+	  case 3:
+	    if (home) {
+		tmpfilename[homelen + 7] = '\0';
+	    }
+	    break;
+
+	  case 4:
+	    if (!filename) {
+		home = getenv ("DECW$USER_DEFAULTS");
+		if (home) {
+		    homelen = strlen (home);
+		    cp = tmpfilename;
+		    (void) sprintf (tmpfilename, "%stwm.rc_%d",
+				    home, Scr->screen);
+		    break;
+		}
+	    }
+	    continue;
+
+	  case 5:
+	    if (home) {
+		tmpfilename[homelen + 6] = '\0';
+	    }
+	    break;
+#else
+	  case 0:			/* -f filename.# */
+	    if (filename) {
+		cp = tmpfilename;
+		(void) sprintf (tmpfilename, "%s.%d", filename, Scr->screen);
+	    }
+	    else cp = filename;
+	    break;
+
+	  case 1:			/* -f filename */
+	    cp = filename;
+	    break;
+
+	  case 2:			/* ~/.ctwmrc.screennum */
 	    if (!filename) {
 		home = getenv ("HOME");
 		if (home) {
@@ -173,13 +284,13 @@ int ParseTwmrc (filename)
 	    }
 	    continue;
 
-	  case 2:			/* ~/.ctwmrc */
+	  case 3:			/* ~/.ctwmrc */
 	    if (home) {
 		tmpfilename[homelen + 8] = '\0';
 	    }
 	    break;
 
-	  case 3:			/* ~/.twmrc.screennum */
+	  case 4:			/* ~/.twmrc.screennum */
 	    if (!filename) {
 		home = getenv ("HOME");
 		if (home) {
@@ -192,13 +303,14 @@ int ParseTwmrc (filename)
 	    }
 	    continue;
 
-	  case 4:			/* ~/.twmrc */
+	  case 5:			/* ~/.twmrc */
 	    if (home) {
 		tmpfilename[homelen + 7] = '\0'; /* C.L. */
 	    }
 	    break;
+#endif
 
-	  case 5:			/* system.twmrc */
+	  case 6:			/* system.twmrc */
 	    cp = SYSTEM_INIT_FILE;
 	    break;
 	}
@@ -219,7 +331,7 @@ int ParseTwmrc (filename)
 #endif
 	int status;
 
-	if (filename && cp != filename) {
+	if (filename && strncmp (cp, filename, strlen (filename))) {
 	    fprintf (stderr,
 		     "%s:  unable to open twmrc file %s, using %s instead\n",
 		     ProgramName, filename, cp);
@@ -444,7 +556,6 @@ typedef struct _TwmKeyword {
 #define kw0_RestartPreviousState	17
 #define kw0_ClientBorderWidth		18
 #define kw0_NoTitleFocus		19
-#define kw0_RandomPlacement		20
 #define kw0_DecorateTransients		21
 #define kw0_ShowIconManager		22
 #define kw0_NoCaseSensitive		23
@@ -459,8 +570,20 @@ typedef struct _TwmKeyword {
 #define kw0_Use3DMenus			33
 #define kw0_Use3DTitles			34
 #define kw0_Use3DIconManagers		35
-#define kw0_SunkFocusWindowTitle	36
-#define kw0_BeNiceToColormap		37
+#define kw0_Use3DBorders		36
+#define kw0_SunkFocusWindowTitle	37
+#define kw0_BeNiceToColormap		38
+#define kw0_SmartIconify		39
+#define kw0_WarpRingOnScreen		40
+#define kw0_NoIconManagerFocus		41
+#define kw0_StayUpMenus			42
+#define kw0_ClickToFocus		43
+#define kw0_BorderResizeCursors		44
+#define kw0_ReallyMoveInWorkspaceManager 45
+#define kw0_ShowWinWhenMovingInWmgr	46
+#define kw0_Use3DWMap			47
+#define kw0_ReverseCurrentWorkspace	48
+#define kw0_DontWarpCursorInWMap	49
 
 #define kws_UsePPosition		1
 #define kws_IconFont			2
@@ -472,6 +595,15 @@ typedef struct _TwmKeyword {
 #define kws_IconDirectory		8
 #define kws_MaxWindowSize		9
 #define kws_PixmapDirectory		10
+#define kws_RandomPlacement		11
+#define kws_IconJustification		12
+#define kws_TitleJustification		13
+#define kws_IconRegionJustification	14
+#define kws_IconRegionAlignement	15
+#ifdef SOUNDS
+#define kws_SoundHost			16
+#endif
+#define kws_WMgrButtonStyle		17
 
 #define kwn_ConstrainedMoveTime		1
 #define kwn_MoveDelta			2
@@ -490,6 +622,17 @@ typedef struct _TwmKeyword {
 #define kwn_WMgrHorizButtonIndent	15
 #define kwn_ClearShadowContrast		16
 #define kwn_DarkShadowContrast		17
+#define kwn_WMgrButtonShadowDepth	18
+#define kwn_MaxIconTitleWidth		19
+#define kwn_AnimationSpeed		20
+#define kwn_ThreeDBorderWidth		21
+#define kwn_MoveOffResistance		22
+#define kwn_BorderShadowDepth		23
+#define kwn_TitleShadowDepth		24
+#define kwn_TitleButtonShadowDepth	25
+#define kwn_MenuShadowDepth		26
+#define kwn_IconManagerShadowDepth	27
+#define kwn_MovePackResistance		28
 
 #define kwcl_BorderColor		1
 #define kwcl_IconManagerHighlight	2
@@ -513,19 +656,25 @@ typedef struct _TwmKeyword {
 #define kwc_MenuTitleBackground		6
 #define kwc_MenuShadowColor		7
 
-
 /*
  * The following is sorted alphabetically according to name (which must be
  * in lowercase and only contain the letters a-z).  It is fed to a binary
  * search to parse keywords.
  */
 static TwmKeyword keytable[] = { 
+    { "a",			ALTER, 0 },
     { "all",			ALL, 0 },
+    { "alter",			ALTER, 0 },
+    { "alwaysontop",		ALWAYS_ON_TOP, 0 },
+    { "alwaysshowwindowwhenmovingfromworkspacemanager", KEYWORD, kw0_ShowWinWhenMovingInWmgr },
+    { "animationspeed",		NKEYWORD, kwn_AnimationSpeed },
     { "autooccupy",		KEYWORD, kw0_AutoOccupy },
     { "autoraise",		AUTO_RAISE, 0 },
     { "autorelativeresize",	KEYWORD, kw0_AutoRelativeResize },
     { "benicetocolormap",	KEYWORD, kw0_BeNiceToColormap },
     { "bordercolor",		CLKEYWORD, kwcl_BorderColor },
+    { "borderresizecursors",	KEYWORD, kw0_BorderResizeCursors },
+    { "bordershadowdepth",	NKEYWORD, kwn_BorderShadowDepth },
     { "bordertilebackground",	CLKEYWORD, kwcl_BorderTileBackground },
     { "bordertileforeground",	CLKEYWORD, kwcl_BorderTileForeground },
     { "borderwidth",		NKEYWORD, kwn_BorderWidth },
@@ -533,7 +682,9 @@ static TwmKeyword keytable[] = {
     { "buttonindent",		NKEYWORD, kwn_ButtonIndent },
     { "c",			CONTROL, 0 },
     { "center",			JKEYWORD, J_CENTER },
+    { "changeworkspacefunction", CHANGE_WORKSPACE_FUNCTION, 0 },
     { "clearshadowcontrast",	NKEYWORD, kwn_ClearShadowContrast },
+    { "clicktofocus",		KEYWORD, kw0_ClickToFocus },
     { "clientborderwidth",	KEYWORD, kw0_ClientBorderWidth },
     { "color",			COLOR, 0 },
     { "constrainedmovetime",	NKEYWORD, kwn_ConstrainedMoveTime },
@@ -544,13 +695,19 @@ static TwmKeyword keytable[] = {
     { "defaultbackground",	CKEYWORD, kwc_DefaultBackground },
     { "defaultforeground",	CKEYWORD, kwc_DefaultForeground },
     { "defaultfunction",	DEFAULT_FUNCTION, 0 },
+    { "deiconifyfunction",	DEICONIFY_FUNCTION, 0 },
     { "destroy",		KILL, 0 },
     { "donticonifybyunmapping",	DONT_ICONIFY_BY_UNMAPPING, 0 },
     { "dontmoveoff",		KEYWORD, kw0_DontMoveOff },
     { "dontpaintrootwindow",	KEYWORD, kw0_DontPaintRootWindow },
     { "dontsqueezetitle",	DONT_SQUEEZE_TITLE, 0 },
+    { "dontwarpcursorinwmap",	KEYWORD, kw0_DontWarpCursorInWMap },
     { "east",			DKEYWORD, D_EAST },
     { "f",			FRAME, 0 },
+    { "f.addtoworkspace",	FSKEYWORD, F_ADDTOWORKSPACE },
+    { "f.adoptwindow",		FKEYWORD, F_ADOPTWINDOW },
+    { "f.altcontext",		FKEYWORD, F_ALTCONTEXT },
+    { "f.altkeymap",		FSKEYWORD, F_ALTKEYMAP },
     { "f.autoraise",		FKEYWORD, F_AUTORAISE },
     { "f.backiconmgr",		FKEYWORD, F_BACKICONMGR },
     { "f.beep",			FKEYWORD, F_BEEP },
@@ -565,6 +722,7 @@ static TwmKeyword keytable[] = {
     { "f.deltastop",		FKEYWORD, F_DELTASTOP },
     { "f.destroy",		FKEYWORD, F_DESTROY },
     { "f.downiconmgr",		FKEYWORD, F_DOWNICONMGR },
+    { "f.downworkspace",	FKEYWORD, F_DOWNWORKSPACE },
     { "f.exec",			FSKEYWORD, F_EXEC },
     { "f.file",			FSKEYWORD, F_FILE },
     { "f.focus",		FKEYWORD, F_FOCUS },
@@ -582,11 +740,14 @@ static TwmKeyword keytable[] = {
     { "f.iconify",		FKEYWORD, F_ICONIFY },
     { "f.identify",		FKEYWORD, F_IDENTIFY },
     { "f.lefticonmgr",		FKEYWORD, F_LEFTICONMGR },
+    { "f.leftworkspace",	FKEYWORD, F_LEFTWORKSPACE },
     { "f.leftzoom",		FKEYWORD, F_LEFTZOOM },
     { "f.lower",		FKEYWORD, F_LOWER },
     { "f.menu",			FSKEYWORD, F_MENU },
     { "f.move",			FKEYWORD, F_MOVE },
     { "f.movemenu",		FKEYWORD, F_MOVEMENU },
+    { "f.movepack",		FKEYWORD, F_MOVEPACK },
+    { "f.movepush",		FKEYWORD, F_MOVEPUSH },
     { "f.nexticonmgr",		FKEYWORD, F_NEXTICONMGR },
     { "f.nextworkspace",	FKEYWORD, F_NEXTWORKSPACE },
     { "f.nop",			FKEYWORD, F_NOP },
@@ -597,11 +758,17 @@ static TwmKeyword keytable[] = {
     { "f.prevworkspace",	FKEYWORD, F_PREVWORKSPACE },
     { "f.quit",			FKEYWORD, F_QUIT },
     { "f.raise",		FKEYWORD, F_RAISE },
+    { "f.raiseicons",		FKEYWORD, F_RAISEICONS },
     { "f.raiselower",		FKEYWORD, F_RAISELOWER },
     { "f.refresh",		FKEYWORD, F_REFRESH },
+    { "f.removefromworkspace",	FSKEYWORD, F_REMOVEFROMWORKSPACE },
+#ifdef SOUNDS
+    { "f.rereadsounds",		FKEYWORD, F_REREADSOUNDS },
+#endif
     { "f.resize",		FKEYWORD, F_RESIZE },
     { "f.restart",		FKEYWORD, F_RESTART },
     { "f.righticonmgr",		FKEYWORD, F_RIGHTICONMGR },
+    { "f.rightworkspace",	FKEYWORD, F_RIGHTWORKSPACE },
     { "f.rightzoom",		FKEYWORD, F_RIGHTZOOM },
     { "f.saveyourself",		FKEYWORD, F_SAVEYOURSELF },
     { "f.separator",		FKEYWORD, F_SEPARATOR },
@@ -609,14 +776,24 @@ static TwmKeyword keytable[] = {
     { "f.setmapstate",		FKEYWORD, F_SETMAPSTATE },
     { "f.showiconmgr",		FKEYWORD, F_SHOWLIST },
     { "f.showworkspacemgr",	FKEYWORD, F_SHOWWORKMGR },
+    { "f.slowdownanimation",	FKEYWORD, F_SLOWDOWNANIMATION },
     { "f.sorticonmgr",		FKEYWORD, F_SORTICONMGR },
     { "f.source",		FSKEYWORD, F_BEEP },  /* XXX - don't work */
+    { "f.speedupanimation",	FKEYWORD, F_SPEEDUPANIMATION },
+    { "f.startanimation",	FKEYWORD, F_STARTANIMATION },
+    { "f.stopanimation",	FKEYWORD, F_STOPANIMATION },
     { "f.title",		FKEYWORD, F_TITLE },
+    { "f.toggleoccupation",	FSKEYWORD, F_TOGGLEOCCUPATION },
+#ifdef SOUNDS
+    { "f.togglesound",		FKEYWORD, F_TOGGLESOUND },
+#endif
     { "f.togglestate",		FKEYWORD, F_TOGGLESTATE },
     { "f.topzoom",		FKEYWORD, F_TOPZOOM },
+    { "f.trace",		FSKEYWORD, F_TRACE },
     { "f.twmrc",		FKEYWORD, F_RESTART },
     { "f.unfocus",		FKEYWORD, F_UNFOCUS },
     { "f.upiconmgr",		FKEYWORD, F_UPICONMGR },
+    { "f.upworkspace",		FKEYWORD, F_UPWORKSPACE },
     { "f.vanish",		FKEYWORD, F_VANISH },
     { "f.version",		FKEYWORD, F_VERSION },
     { "f.vlzoom",		FKEYWORD, F_LEFTZOOM },
@@ -641,6 +818,8 @@ static TwmKeyword keytable[] = {
     { "iconfont",		SKEYWORD, kws_IconFont },
     { "iconforeground",		CLKEYWORD, kwcl_IconForeground },
     { "iconifybyunmapping",	ICONIFY_BY_UNMAPPING, 0 },
+    { "iconifyfunction",	ICONIFY_FUNCTION, 0 },
+    { "iconjustification",	SKEYWORD, kws_IconJustification },
     { "iconmanagerbackground",	CLKEYWORD, kwcl_IconManagerBackground },
     { "iconmanagerdontshow",	ICONMGR_NOSHOW, 0 },
     { "iconmanagerfont",	SKEYWORD, kws_IconManagerFont },
@@ -648,9 +827,12 @@ static TwmKeyword keytable[] = {
     { "iconmanagergeometry",	ICONMGR_GEOMETRY, 0 },
     { "iconmanagerhighlight",	CLKEYWORD, kwcl_IconManagerHighlight },
     { "iconmanagers",		ICONMGRS, 0 },
+    { "iconmanagershadowdepth",	NKEYWORD, kwn_IconManagerShadowDepth },
     { "iconmanagershow",	ICONMGR_SHOW, 0 },
     { "iconmgr",		ICONMGR, 0 },
     { "iconregion",		ICON_REGION, 0 },
+    { "iconregionalignement",	SKEYWORD, kws_IconRegionAlignement },
+    { "iconregionjustification",SKEYWORD, kws_IconRegionJustification },
     { "icons",			ICONS, 0 },
     { "interpolatemenucolors",	KEYWORD, kw0_InterpolateMenuColors },
     { "l",			LOCK, 0 },
@@ -663,12 +845,14 @@ static TwmKeyword keytable[] = {
     { "mapwindowcurrentworkspace", MAPWINDOWCURRENTWORKSPACE, 0},
     { "mapwindowdefaultworkspace", MAPWINDOWDEFAULTWORKSPACE, 0},
     { "mapwindowforeground",	CLKEYWORD, kwcl_MapWindowForeground },
+    { "maxicontitlewidth",	NKEYWORD, kwn_MaxIconTitleWidth },
     { "maxwindowsize",		SKEYWORD, kws_MaxWindowSize },
     { "menu",			MENU, 0 },
     { "menubackground",		CKEYWORD, kwc_MenuBackground },
     { "menufont",		SKEYWORD, kws_MenuFont },
     { "menuforeground",		CKEYWORD, kwc_MenuForeground },
     { "menushadowcolor",	CKEYWORD, kwc_MenuShadowColor },
+    { "menushadowdepth",	NKEYWORD, kwn_MenuShadowDepth },
     { "menutitlebackground",	CKEYWORD, kwc_MenuTitleBackground },
     { "menutitleforeground",	CKEYWORD, kwc_MenuTitleForeground },
     { "meta",			META, 0 },
@@ -676,11 +860,15 @@ static TwmKeyword keytable[] = {
     { "monochrome",		MONOCHROME, 0 },
     { "move",			MOVE, 0 },
     { "movedelta",		NKEYWORD, kwn_MoveDelta },
+    { "moveoffresistance",	NKEYWORD, kwn_MoveOffResistance },
+    { "movepackresistance",	NKEYWORD, kwn_MovePackResistance },
     { "nobackingstore",		KEYWORD, kw0_NoBackingStore },
+    { "noborder",		NO_BORDER, 0 },
     { "nocasesensitive",	KEYWORD, kw0_NoCaseSensitive },
     { "nodefaults",		KEYWORD, kw0_NoDefaults },
     { "nograbserver",		KEYWORD, kw0_NoGrabServer },
     { "nohighlight",		NO_HILITE, 0 },
+    { "noiconmanagerfocus",	KEYWORD, kw0_NoIconManagerFocus },
     { "noiconmanagers",		KEYWORD, kw0_NoIconManagers },
     { "noicontitle",		NO_ICON_TITLE, 0  },
     { "nomenushadows",		KEYWORD, kw0_NoMenuShadows },
@@ -708,10 +896,12 @@ static TwmKeyword keytable[] = {
     { "pixmaps",		PIXMAPS, 0 },
     { "r",			ROOT, 0 },
     { "raisedelay",		NKEYWORD, kwn_RaiseDelay },
-    { "randomplacement",	KEYWORD, kw0_RandomPlacement },
+    { "randomplacement",	SKEYWORD, kws_RandomPlacement },
+    { "reallymoveinworkspacemanager",	KEYWORD, kw0_ReallyMoveInWorkspaceManager },
     { "resize",			RESIZE, 0 },
     { "resizefont",		SKEYWORD, kws_ResizeFont },
     { "restartpreviousstate",	KEYWORD, kw0_RestartPreviousState },
+    { "reversecurrentworkspace",KEYWORD, kw0_ReverseCurrentWorkspace },
     { "right",			JKEYWORD, J_RIGHT },
     { "righttitlebutton",	RIGHT_TITLEBUTTON, 0 },
     { "root",			ROOT, 0 },
@@ -721,35 +911,49 @@ static TwmKeyword keytable[] = {
     { "shift",			SHIFT, 0 },
     { "showiconmanager",	KEYWORD, kw0_ShowIconManager },
     { "showworkspacemanager",	KEYWORD, kw0_ShowWorkspaceManager },
+    { "smarticonify",		KEYWORD, kw0_SmartIconify },
     { "sorticonmanager",	KEYWORD, kw0_SortIconManager },
+#ifdef SOUNDS
+    { "soundhost",		SKEYWORD, kws_SoundHost },
+#endif
     { "south",			DKEYWORD, D_SOUTH },
     { "squeezetitle",		SQUEEZE_TITLE, 0 },
     { "starticonified",		START_ICONIFIED, 0 },
     { "startinmapstate",	KEYWORD, kw0_StartInMapState },
+    { "stayupmenus",		KEYWORD, kw0_StayUpMenus },
     { "sunkfocuswindowtitle",	KEYWORD, kw0_SunkFocusWindowTitle },
     { "t",			TITLE, 0 },
+    { "threedborderwidth",	NKEYWORD, kwn_ThreeDBorderWidth },
     { "title",			TITLE, 0 },
     { "titlebackground",	CLKEYWORD, kwcl_TitleBackground },
     { "titlebuttonborderwidth",	NKEYWORD, kwn_TitleButtonBorderWidth },
+    { "titlebuttonshadowdepth",	NKEYWORD, kwn_TitleButtonShadowDepth },
     { "titlefont",		SKEYWORD, kws_TitleFont },
     { "titleforeground",	CLKEYWORD, kwcl_TitleForeground },
     { "titlehighlight",		TITLE_HILITE, 0 },
+    { "titlejustification",	SKEYWORD, kws_TitleJustification },
     { "titlepadding",		NKEYWORD, kwn_TitlePadding },
+    { "titleshadowdepth",	NKEYWORD, kwn_TitleShadowDepth },
     { "transienthasoccupation",	KEYWORD, kw0_TransientHasOccupation },
     { "transientontop",		NKEYWORD, kwn_TransientOnTop },
     { "unknownicon",		SKEYWORD, kws_UnknownIcon },
     { "usepposition",		SKEYWORD, kws_UsePPosition },
+    { "usethreedborders",	KEYWORD, kw0_Use3DBorders },
     { "usethreediconmanagers",	KEYWORD, kw0_Use3DIconManagers },
     { "usethreedmenus",		KEYWORD, kw0_Use3DMenus },
     { "usethreedtitles",	KEYWORD, kw0_Use3DTitles },
+    { "usethreedwmap",		KEYWORD, kw0_Use3DWMap },
     { "w",			WINDOW, 0 },
     { "wait",			WAITC, 0 },
     { "warpcursor",		WARP_CURSOR, 0 },
+    { "warpringonscreen",	KEYWORD, kw0_WarpRingOnScreen },
     { "warpunmapped",		KEYWORD, kw0_WarpUnmapped },
     { "west",			DKEYWORD, D_WEST },
     { "window",			WINDOW, 0 },
     { "windowfunction",		WINDOW_FUNCTION, 0 },
     { "windowring",		WINDOW_RING, 0 },
+    { "wmgrbuttonshadowdepth",	NKEYWORD, kwn_WMgrButtonShadowDepth },
+    { "wmgrbuttonstyle",	SKEYWORD, kws_WMgrButtonStyle },
     { "wmgrhorizbuttonindent",	NKEYWORD, kwn_WMgrHorizButtonIndent },
     { "wmgrvertbuttonindent",	NKEYWORD, kwn_WMgrVertButtonIndent },
     { "workspace", 		WORKSPACE, 0 },
@@ -865,11 +1069,7 @@ int do_single_keyword (keyword)
 	return 1;
 
       case kw0_NoTitleFocus:
-	Scr->TitleFocus = FALSE;
-	return 1;
-
-      case kw0_RandomPlacement:
-	Scr->RandomPlacement = TRUE;
+	Scr->TitleFocus = TRUE /*FALSE*/;
 	return 1;
 
       case kw0_DecorateTransients:
@@ -904,6 +1104,10 @@ int do_single_keyword (keyword)
 	Scr->DontPaintRootWindow = TRUE;
 	return 1;
 
+      case kw0_Use3DBorders:
+	Scr->use3Dborders = TRUE;
+	return 1;
+
       case kw0_Use3DIconManagers:
 	Scr->use3Diconmanagers = TRUE;
 	return 1;
@@ -916,6 +1120,10 @@ int do_single_keyword (keyword)
 	Scr->use3Dtitles = TRUE;
 	return 1;
 
+      case kw0_Use3DWMap:
+	Scr->use3Dwmap = TRUE;
+	return 1;
+
       case kw0_SunkFocusWindowTitle:
 	Scr->SunkFocusWindowTitle = TRUE;
 	return 1;
@@ -924,6 +1132,14 @@ int do_single_keyword (keyword)
 	Scr->BeNiceToColormap = TRUE;
 	return 1;
 
+      case kw0_SmartIconify:
+	Scr->SmartIconify = TRUE;
+	return 1;
+
+      case kw0_BorderResizeCursors:
+	Scr->BorderCursors = TRUE;
+	return 1;
+
       case kw0_NoCaseSensitive:
 	Scr->CaseSensitive = FALSE;
 	return 1;
@@ -935,8 +1151,39 @@ int do_single_keyword (keyword)
       case kw0_WarpUnmapped:
 	Scr->WarpUnmapped = TRUE;
 	return 1;
-    }
 
+      case kw0_WarpRingOnScreen:
+	Scr->WarpRingAnyWhere = FALSE;
+	return 1;
+
+      case kw0_NoIconManagerFocus:
+	Scr->IconManagerFocus = FALSE;
+	return 1;
+
+      case kw0_StayUpMenus:
+	Scr->StayUpMenus = TRUE;
+	return 1;
+
+      case kw0_ClickToFocus:
+	Scr->ClickToFocus = TRUE;
+	return 1;
+
+      case kw0_ReallyMoveInWorkspaceManager:
+	Scr->ReallyMoveInWorkspaceManager = TRUE;
+	return 1;
+
+      case kw0_ShowWinWhenMovingInWmgr:
+	Scr->ShowWinWhenMovingInWmgr = TRUE;
+	return 1;
+
+      case kw0_ReverseCurrentWorkspace:
+	Scr->ReverseCurrentWorkspace = TRUE;
+	return 1;
+
+      case kw0_DontWarpCursorInWMap:
+	Scr->DontWarpCursorInWMap = TRUE;
+	return 1;
+    }
     return 0;
 }
 
@@ -946,6 +1193,18 @@ int do_string_keyword (keyword, s)
     char *s;
 {
     switch (keyword) {
+       case kws_RandomPlacement:
+ 	{
+ 	    int rp = ParseRandomPlacement (s);
+ 	    if (rp < 0) {
+ 		twmrc_error_prefix();
+ 		fprintf (stderr,
+ 			 "ignoring invalid RandomPlacement argument \"%s\"\n", s);
+ 	    } else {
+ 		Scr->RandomPlacement = rp;
+ 	    }
+ 	    return 1;
+ 	}
       case kws_UsePPosition:
 	{ 
 	    int ppos = ParseUsePPosition (s);
@@ -1007,8 +1266,80 @@ int do_string_keyword (keyword, s)
 	Scr->MaxWindowWidth = JunkWidth;
 	Scr->MaxWindowHeight = JunkHeight;
 	return 1;
-    }
 
+      case kws_IconJustification:
+	{
+	    int just = ParseJustification (s);
+
+ 	    if ((just < 0) || (just == J_BORDER)) {
+ 		twmrc_error_prefix();
+ 		fprintf (stderr,
+ 			 "ignoring invalid IconJustification argument \"%s\"\n", s);
+ 	    } else {
+ 		Scr->IconJustification = just;
+ 	    }
+ 	    return 1;
+ 	}
+      case kws_IconRegionJustification:
+	{
+	    int just = ParseJustification (s);
+
+ 	    if (just < 0) {
+ 		twmrc_error_prefix();
+ 		fprintf (stderr,
+ 			 "ignoring invalid IconRegionJustification argument \"%s\"\n", s);
+ 	    } else {
+ 		Scr->IconRegionJustification = just;
+ 	    }
+ 	    return 1;
+ 	}
+      case kws_IconRegionAlignement:
+	{
+	    int just = ParseAlignement (s);
+
+ 	    if (just < 0) {
+ 		twmrc_error_prefix();
+ 		fprintf (stderr,
+ 			 "ignoring invalid IconRegionAlignement argument \"%s\"\n", s);
+ 	    } else {
+ 		Scr->IconRegionAlignement = just;
+ 	    }
+ 	    return 1;
+ 	}
+
+      case kws_TitleJustification:
+	{
+	    int just = ParseJustification (s);
+
+ 	    if ((just < 0) || (just == J_BORDER)) {
+ 		twmrc_error_prefix();
+ 		fprintf (stderr,
+ 			 "ignoring invalid TitleJustification argument \"%s\"\n", s);
+ 	    } else {
+ 		Scr->TitleJustification = just;
+ 	    }
+ 	    return 1;
+ 	}
+#ifdef SOUNDS
+      case kws_SoundHost:
+        if (Scr->FirstTime) set_sound_host(s);
+        return 1;
+#endif
+		
+      case kws_WMgrButtonStyle:
+	{
+	    int style = ParseButtonStyle (s);
+
+ 	    if (style < 0) {
+ 		twmrc_error_prefix();
+ 		fprintf (stderr,
+ 			 "ignoring invalid WMgrButtonStyle argument \"%s\"\n", s);
+ 	    } else {
+ 		Scr->workSpaceMgr.workspaceWindow.buttonStyle = style;
+ 	    }
+ 	    return 1;
+	}
+    }
     return 0;
 }
 
@@ -1026,6 +1357,15 @@ int do_number_keyword (keyword, num)
 	Scr->MoveDelta = num;
 	return 1;
 
+      case kwn_MoveOffResistance:
+	Scr->MoveOffResistance = num;
+	return 1;
+
+      case kwn_MovePackResistance:
+	if (num < 0) num = 20;
+	Scr->MovePackResistance = num;
+	return 1;
+
       case kwn_XorValue:
 	if (Scr->FirstTime) Scr->XORvalue = num;
 	return 1;
@@ -1044,6 +1384,10 @@ int do_number_keyword (keyword, num)
 	if (Scr->FirstTime) Scr->ButtonIndent = num;
 	return 1;
 
+      case kwn_ThreeDBorderWidth:
+	if (Scr->FirstTime) Scr->ThreeDBorderWidth = num;
+	return 1;
+
       case kwn_BorderWidth:
 	if (Scr->FirstTime) Scr->BorderWidth = num;
 	return 1;
@@ -1082,6 +1426,15 @@ int do_number_keyword (keyword, num)
 	if (Scr->WMgrHorizButtonIndent < 0) Scr->WMgrHorizButtonIndent = 0;
 	return 1;
 
+      case kwn_WMgrButtonShadowDepth:
+	if (Scr->FirstTime) Scr->WMgrButtonShadowDepth = num;
+	if (Scr->WMgrButtonShadowDepth < 1) Scr->WMgrButtonShadowDepth = 1;
+	return 1;
+
+      case kwn_MaxIconTitleWidth:
+	if (Scr->FirstTime) Scr->MaxIconTitleWidth = num;
+	return 1;
+
       case kwn_ClearShadowContrast:
 	if (Scr->FirstTime) Scr->ClearShadowContrast = num;
 	if (Scr->ClearShadowContrast <   0) Scr->ClearShadowContrast =   0;
@@ -1093,6 +1446,37 @@ int do_number_keyword (keyword, num)
 	if (Scr->DarkShadowContrast <   0) Scr->DarkShadowContrast =   0;
 	if (Scr->DarkShadowContrast > 100) Scr->DarkShadowContrast = 100;
 	return 1;
+
+      case kwn_AnimationSpeed:
+	if (num < 0) num = 0;
+	SetAnimationSpeed (num);
+	return 1;
+
+      case kwn_BorderShadowDepth:
+	if (Scr->FirstTime) Scr->BorderShadowDepth = num;
+	if (Scr->BorderShadowDepth < 0) Scr->BorderShadowDepth = 2;
+	return 1;
+
+      case kwn_TitleButtonShadowDepth:
+	if (Scr->FirstTime) Scr->TitleButtonShadowDepth = num;
+	if (Scr->TitleButtonShadowDepth < 0) Scr->TitleButtonShadowDepth = 2;
+	return 1;
+
+      case kwn_TitleShadowDepth:
+	if (Scr->FirstTime) Scr->TitleShadowDepth = num;
+	if (Scr->TitleShadowDepth < 0) Scr->TitleShadowDepth = 2;
+	return 1;
+
+      case kwn_IconManagerShadowDepth:
+	if (Scr->FirstTime) Scr->IconManagerShadowDepth = num;
+	if (Scr->IconManagerShadowDepth < 0) Scr->IconManagerShadowDepth = 2;
+	return 1;
+
+      case kwn_MenuShadowDepth:
+	if (Scr->FirstTime) Scr->MenuShadowDepth = num;
+	if (Scr->MenuShadowDepth < 0) Scr->MenuShadowDepth = 2;
+	return 1;
+
     }
 
     return 0;
@@ -1105,7 +1489,7 @@ name_list **do_colorlist_keyword (keyword, colormode, s)
 {
     switch (keyword) {
       case kwcl_BorderColor:
-	GetColor (colormode, &Scr->BorderColor, s);
+	GetColor (colormode, &Scr->BorderColorC.back, s);
 	return &Scr->BorderColorL;
 
       case kwcl_IconManagerHighlight:
@@ -1150,10 +1534,12 @@ name_list **do_colorlist_keyword (keyword, colormode, s)
 
       case kwcl_MapWindowBackground:
 	GetColor (colormode, &Scr->workSpaceMgr.workspaceWindow.windowcp.back, s);
+	Scr->workSpaceMgr.workspaceWindow.windowcpgiven = True;
 	return &Scr->workSpaceMgr.workspaceWindow.windowBackgroundL;
 
       case kwcl_MapWindowForeground:
 	GetColor (colormode, &Scr->workSpaceMgr.workspaceWindow.windowcp.fore, s);
+	Scr->workSpaceMgr.workspaceWindow.windowcpgiven = True;
 	return &Scr->workSpaceMgr.workspaceWindow.windowForegroundL;
     }
     return NULL;
@@ -1268,7 +1654,7 @@ void assign_var_savecolor()
   while (cp != NULL) {
     switch (cp->i) {
     case kwcl_BorderColor:
-      put_pixel_on_root(Scr->BorderColor);
+      put_pixel_on_root(Scr->BorderColorC.back);
       break;
     case kwcl_IconManagerHighlight:
       put_pixel_on_root(Scr->IconManagerHighlight);
@@ -1315,23 +1701,78 @@ void assign_var_savecolor()
   }
 }
 
+static int ParseRandomPlacement (s)
+    register char *s;
+{
+    if (strlen (s) == 0) return RP_ALL;
+    if (strcmp (s, "default") == 0) return RP_ALL;
+    XmuCopyISOLatin1Lowered (s, s);
+
+    if (strcmp (s,      "off") == 0) return RP_OFF;
+    if (strcmp (s,       "on") == 0) return RP_ALL;
+    if (strcmp (s,      "all") == 0) return RP_ALL;
+    if (strcmp (s, "unmapped") == 0) return RP_UNMAPPED;
+    return (-1);
+}
+
+static int ParseJustification (s)
+    register char *s;
+{
+    if (strlen (s) == 0) return (-1);
+    if (strcmp (s, "default") == 0) return J_CENTER;
+    XmuCopyISOLatin1Lowered (s, s);
+
+    if (strcmp (s, "default") == 0) return J_CENTER;
+    if (strcmp (s,    "left") == 0) return J_LEFT;
+    if (strcmp (s,  "center") == 0) return J_CENTER;
+    if (strcmp (s,   "right") == 0) return J_RIGHT;
+    if (strcmp (s,  "border") == 0) return J_BORDER;
+    return (-1);
+}
+
+static int ParseAlignement (s)
+    register char *s;
+{
+    if (strlen (s) == 0) return (-1);
+    if (strcmp (s, "default") == 0) return J_CENTER;
+    XmuCopyISOLatin1Lowered (s, s);
+
+    if (strcmp (s, "default") == 0) return J_CENTER;
+    if (strcmp (s,     "top") == 0) return J_TOP;
+    if (strcmp (s,  "center") == 0) return J_CENTER;
+    if (strcmp (s,  "bottom") == 0) return J_BOTTOM;
+    if (strcmp (s,  "border") == 0) return J_BORDER;
+    return (-1);
+}
+
 static int ParseUsePPosition (s)
     register char *s;
 {
+    if (strlen (s) == 0) return (-1);
+    if (strcmp (s,  "default") == 0) return PPOS_OFF;
     XmuCopyISOLatin1Lowered (s, s);
 
-    if (strcmp (s, "off") == 0) {
-	return PPOS_OFF;
-    } else if (strcmp (s, "on") == 0) {
-	return PPOS_ON;
-    } else if (strcmp (s, "non-zero") == 0 ||
-	       strcmp (s, "nonzero") == 0) {
-	return PPOS_NON_ZERO;
-    }
-
-    return -1;
+    if (strcmp (s,  "default") == 0) return PPOS_OFF;
+    if (strcmp (s,      "off") == 0) return PPOS_OFF;
+    if (strcmp (s,       "on") == 0) return PPOS_ON;
+    if (strcmp (s, "non-zero") == 0) return PPOS_NON_ZERO;
+    if (strcmp (s,  "nonzero") == 0) return PPOS_NON_ZERO;
+    return (-1);
 }
 
+static int ParseButtonStyle (s)
+    register char *s;
+{
+    if (strlen (s) == 0) return (-1);
+    if (strcmp (s,  "default") == 0) return STYLE_NORMAL;
+    XmuCopyISOLatin1Lowered (s, s);
+
+    if (strcmp (s, "normal") == 0) return STYLE_NORMAL;
+    if (strcmp (s, "style1") == 0) return STYLE_STYLE1;
+    if (strcmp (s, "style2") == 0) return STYLE_STYLE2;
+    if (strcmp (s, "style3") == 0) return STYLE_STYLE3;
+    return (-1);
+}
 
 do_squeeze_entry (list, name, justify, num, denom)
     name_list **list;			/* squeeze or dont-squeeze list */
@@ -1345,13 +1786,13 @@ do_squeeze_entry (list, name, justify, num, denom)
     if (denom < 0) {
 	twmrc_error_prefix();
 	fprintf (stderr, "negative SqueezeTitle denominator %d\n", denom);
-	return;
+	return (1);
     }
     if (absnum > denom && denom != 0) {
 	twmrc_error_prefix();
 	fprintf (stderr, "SqueezeTitle fraction %d/%d outside window\n",
 		 num, denom);
-	return;
+	return (1);
     }
     if (denom == 1) {
 	twmrc_error_prefix();
@@ -1369,13 +1810,14 @@ do_squeeze_entry (list, name, justify, num, denom)
 	    twmrc_error_prefix();
 	    fprintf (stderr, "unable to allocate %d bytes for squeeze info\n",
 		     sizeof(SqueezeInfo));
-	    return;
+	    return (1);
 	}
 	sinfo->justify = justify;
 	sinfo->num = num;
 	sinfo->denom = denom;
 	AddToList (list, name, (char *) sinfo);
     }
+    return (0);
 }
 
 #ifdef USEM4
@@ -1414,7 +1856,7 @@ FILE *fraw;
         }
         /* Parent */
         close(fids[1]);
-        return(fdopen(fids[0], "r"));
+        return ((FILE*)fdopen(fids[0], "r"));
 }
 
 /* Code taken and munged from xrdb.c */
@@ -1429,6 +1871,7 @@ char *name, *def;
         static int maxsize = 0;
         int n, nl;
 
+	if (def == NULL) return ("");     /* XXX JWS: prevent segfaults */
         /* The char * storage only lasts for 1 call... */
         if ((n = EXTRA + ((nl = strlen(name)) +  strlen(def))) > maxsize) {
 		maxsize = n;
@@ -1508,22 +1951,20 @@ char *host;
         static char tmp_name[] = "/tmp/twmrcXXXXXX";
         int fd;
         FILE *tmpf;
+	char *user;
 
-        fd = mkstemp(tmp_name);                 /* I *hope* mkstemp exists, beca
-use */
-                                                                        /* I tri
-ed to find the "portable" */
-                                                                        /* mktmp
-... */
+        fd = mkstemp(tmp_name);		/* I *hope* mkstemp exists, because */
+					/* I tried to find the "portable" */
+					/* mktmp... */
         if (fd < 0) {
                 perror("mkstemp failed in m4_defs");
                 exit(377);
         }
-        tmpf = fdopen(fd, "w+");
+        tmpf = (FILE*) fdopen(fd, "w+");
         XmuGetHostname(client, MAXHOSTNAME);
         hostname = gethostbyname(client);
         strcpy(server, XDisplayName(host));
-        colon = index(server, ':');
+        colon = strchr(server, ':');
         if (colon != NULL) *colon = '\0';
         if ((server[0] == '\0') || (!strcmp(server, "unix")))
                 strcpy(server, client); /* must be connected to :0 or unix:0 */
@@ -1535,7 +1976,9 @@ ed to find the "portable" */
                 fputs(MkDef("HOSTNAME", hostname->h_name), tmpf);
         else
                 fputs(MkDef("HOSTNAME", client), tmpf);
-        fputs(MkDef("USER", getenv("USER")), tmpf);
+
+	if (!(user=getenv("USER")) && !(user=getenv("LOGNAME"))) user = "unknown";
+        fputs(MkDef("USER", user), tmpf);
         fputs(MkDef("HOME", getenv("HOME")), tmpf);
         fputs(MkNum("VERSION", ProtocolVersion(display)), tmpf);
         fputs(MkNum("REVISION", ProtocolRevision(display)), tmpf);
@@ -1565,6 +2008,9 @@ ed to find the "portable" */
         } else {
                 fputs(MkDef("COLOR", "No"), tmpf);
         }
+#ifdef XPM
+	fputs(MkDef("XPM", "Yes"), tmpf);
+#endif
         if (KeepTmpFile) {
                 fprintf(stderr, "Left file: %s\n", tmp_name);
         } else {
